function MSNSession(client, email) {
  // object-specific
  this.bundle = getStringBundle("chrome://msnmsgr/locale/session.properties");
  this.client = client;

  // user-related
  this.contacts = new Array();
  this.contacts.push(this.client.getContact(email));

  // connection-related
  this.socket = new Socket();
  this.isConnected = false;
  this._TRID = 0;
  this.data = '';

  // session-related
  this.sessionID = 0;
  this.sessionStatus = 0;
  this.dontSend = false;
  this.msgBuffer = new Array();
  this.lastMsgRcvTime = null;

  this.__defineGetter__("TrID", function() {
    this._TRID++;
    return this._TRID;
  });
}

MSNSession.prototype.open = function(host, port, cki) {
  this._initialSend = 'USR ' + this.TrID + ' ' + this.client.passport + ' ' + cki + '\r\n';
  this.socket.connect(host, port, this);
}

MSNSession.prototype.exit = function() {
  if (typeof(this.sendTimeout) != "undefined") {
    clearTimeout(this.sendTimeout);
  }
  if (this.isConnected == true) {
    this.socket.sendData('OUT\r\n');
    this.socket.disconnect();
  }
  this.client.unregisterSession(this.sessionID);
}

MSNSession.prototype.close = function() {
  window.close();
}

MSNSession.prototype.onConnect = function() {
  this.isConnected = true;
  this.socket.startAsyncRead();
  this.socket.sendData(this._initialSend);
  delete this._initialSend;
}

MSNSession.prototype.onStreamClose = function(status) {
  this.isConnected = false;
  dump("Stream is closed.\r\n");
}

MSNSession.prototype.onClose = function() {
  dump("Connection is closed.\r\n");
}

MSNSession.prototype.onDataAvailable = function(count) {
  this.data += this.socket.readData(count);
  while (this.data != '') {
    var i = this.data.indexOf('\r\n');
    if (i == -1) { return }
    var data = this.data.substr(0, i);
    this.data = this.data.substr(i+2);
    dump("<<< " + data + '\r\n');
    var param = data.split(" ");
    switch (param[0]) {
      case "USR":
        if (param[2] == "OK") {
          this.invite(this.contacts[0].email);
        }
        break;
      case "CAL":
        if (param[2] == "RINGING") {
          this.sessionID = param[3];
          this.client.registerSession(this.sessionID, this);
        }
        break;
      case "JOI":
        var contact = this.client.getContact(param[1]);
        if (this.contacts.length != 1) {
          this.contacts.push(contact);
          contactsChanged();
        }
        if (this.contacts.length > 1) {
          var joinMsg = this.bundle.GetStringFromName("joinMessage");
          showNotificationMessage(joinMsg.replace("%S", unescape(param[2])));
        }
        this.sessionStatus = 3;
        if (this.msgBuffer.length != 0) {
          for (var i in this.msgBuffer) {
            this.sendIM(this.msgBuffer[i]);
          }
          this.msgBuffer = new Array();
        }
        break;
      case "ANS":
        this.sessionStatus = 3;
        contactsChanged();
        break;
      case "IRO":
        var contact = this.client.getContact(param[4]);
        if (contact.email != this.contacts[0].email) {
          this.contacts.push(contact);
        }
        break;
      case "BYE":
        var contact = this.getContact(param[1]);
        if (this.contacts.length == 1) {
          this.sessionStatus = 1;
        }
        else {
          this.contacts.splice(this.getContactIndex(param[1]), 1);
          var leaveMsg = this.bundle.GetStringFromName("leaveMessage");
          showNotificationMessage(leaveMsg.replace("%S", contact.nickname));
          contactsChanged();
        }
        break;
      case "MSG":
        if (this.data.length >= param[3]) {
          var msgData = this.data.substr(0, param[3]);
          this.data = this.data.substr(param[3]);
          var msgParts = msgData.split('\r\n\r\n', 2);
          var headers = parseHeaders(msgParts[0]);
          msgData = msgParts[1];
          dump(msgData + '\r\n');
          this.messageReceived(unescape(uniConvertTo(param[2])), param[1], headers, msgData);
        }
        else {
          this.data = data + this.data;
          return;
        }
        break;
      case "NAK":
        showNotificationMessage(this.bundle.GetStringFromName("messageNotSent"));
        break;
      case "217":
        showNotificationMessage(this.bundle.GetStringFromName("contactOffline"));
        break;
      default:
        alert("Unexpected command: " + data);
        break;
      }
  }
}

MSNSession.prototype.invite = function(email) {
  this.socket.sendData('CAL ' + this.TrID + ' ' + email + '\r\n');
}

MSNSession.prototype.answer = function(host, port, cki, sessionID) {
  this._initialSend = 'ANS ' + this.TrID + ' ' + this.client.passport + ' ' + cki + ' ' + sessionID + '\r\n';
  this.socket.connect(host, port, this);
  this.sessionID = sessionID;
  this.client.registerSession(this.sessionID, this);
}

MSNSession.prototype.handleTyping = function() {
  if (this.contacts.length == 1) {
    if (this.isConnected == true) {
      if (this.sessionStatus == 1) {
        this.sessionStatus = 2;
        this.invite(this.contacts[0].email);
      }
    }
    else {
      if (this.sessionStatus == 1) {
        this.sessionStatus = 4;
        createNewSession(this.contacts[0].email);
      }
    }
  }
  if (this.isConnected == true && this.sessionStatus == 3 && this.dontSend == false) {
    this.dontSend = true;
    this.sendTyping();
    var self = this;
    this.sendTimeout = setTimeout(function() { self.dontSend = false; }, 5000);
  }
}

MSNSession.prototype.sendTyping = function() {
  var msg = "MIME-Version: 1.0\r\n";
  msg += "Content-Type: text/x-msmsgscontrol\r\n";
  msg += "TypingUser: " + this.client.passport + "\r\n";
  msg += "\r\n";
  msg = "MSG " + this.TrID + " U " + msg.length + "\r\n" + msg;
  this.socket.sendData(msg);
}

MSNSession.prototype.sendIM = function(msgText, mms) {
  if (this.sessionStatus == 3) {
    var payload = "MIME-Version: 1.0\r\n";
    payload += "Content-Type: text/plain; charset=UTF-8\r\n";
    payload += "X-MMS-IM-Format: " + mms + "\r\n";
    payload += '\r\n';
    payload += msgText;
    var msg = "MSG " + this.TrID + " N " + payload.length + "\r\n" + payload;
    this.socket.sendData(msg);
    dump('\r\n');
  }
  else {
    this.msgBuffer.push(msgText);
  }
}

MSNSession.prototype.messageReceived = function(from, fromEmail, headers, body) {
  switch (headers["content-type"]) {
    case "text/plain; charset=UTF-8":
      showIM(from, fromEmail, uniConvertTo(body), headers["x-mms-im-format"], false);
      var contact = this.getContact(fromEmail);
      contact.timeout = 0;
      var d = new Date();
      var currentTime = d.getDate() + "/" + d.getMonth() + "/" + d.getFullYear() + " ";
      currentTime += d.getHours() + ":" + d.getMinutes();
      this.lastMsgRcvTime = currentTime;
      this.updateSB();
      window.getAttention();
      break;
    case "text/x-msmsgscontrol":
      var typinguser = headers["typinguser"];
      var contact = null;
      contact = this.getContact(typinguser);
      contact.timeout = new Date().getTime();
      var self = this;
      setTimeout(function() { self.updateSB(); }, 6000);
      this.updateSB();
      break;
    case "text/x-msmsgsinvite; charset=UTF-8":
      var fields = body.split('\r\n');
      var params = parseFields(fields, ": ");
      if (params["invitation-command"] == "INVITE") {
        this.cancelInvite(params["invitation-cookie"], "REJECT_NOT_INSTALLED");
        showNotificationMessage(this.bundle.GetStringFromName("inviteUnsupported"));
      }
      break;
  }
}

MSNSession.prototype.cancelInvite = function(invCookie, cancelCode) {
  var payload = "MIME-Version: 1.0\r\n";
  payload += "Content-Type: text/x-msmsgsinvite; charset=UTF-8\r\n";
  payload += '\r\n';
  payload += "Invitation-Command: CANCEL\r\n";
  payload += "Invitation-Cookie: " + invCookie + "\r\n";
  payload += "Cancel-Code: " + cancelCode + "\r\n";
  msg = "MSG " + this.TrID + " N " + payload.length + "\r\n" + payload;
  this.socket.sendData(msg);
}

MSNSession.prototype.generateTalkingToString = function() {
  var contactsString = '';
  if (this.contacts.length != 0) {
    contactsString = this.bundle.GetStringFromName("sendMessageToLabel") + "  ";
    for (var i in this.contacts) {
      var contact = this.contacts[i];
      contactsString += contact.nickname + " (" + this.client.bundle.GetStringFromName("StatusMsg" + contact.status) + ") <" + contact.email + ">, ";
    }
    contactsString = contactsString.substr(0, contactsString.length - 2)
  }
  else {
    contactsString = this.bundle.GetStringFromName("toNoOne");
  }
  return contactsString;
}

MSNSession.prototype.updateSB = function() {
  var statusMsg = '';
  var count = 0;
  var t = new Date().getTime();
  for (var i in this.contacts) {
    var contact = this.contacts[i];
    if (t <= contact.timeout + 5000) {
      count++;
      statusMsg += contact.nickname + ", ";
    }
  }
  switch (count) {
    case 0:
      if (this.lastMsgRcvTime != null) {
        statusMsg = this.bundle.GetStringFromName("lastMessageReceived").replace("%T", this.lastMsgRcvTime);
      }
      else {
        statusMsg = this.bundle.GetStringFromName("noMessagesReceived");
      }
      break;
    case 1:
      var typingMsg = this.bundle.GetStringFromName("isTyping");
      statusMsg = typingMsg.replace("%U", statusMsg.substr(0, statusMsg.length - 2));
      break;
    default:
      var typingMsg = this.bundle.GetStringFromName("areTyping");
      statusMsg = typingMsg.replace("%U", statusMsg.substr(0, statusMsg.length - 2));
      break;
  }
  setStatus(statusMsg);
}

MSNSession.prototype.contactChanged = function(contact) {
  for (var i = 0; i < this.contacts.length; i++) {
    var sessionContact = this.contacts[i];
    if (sessionContact.email == contact.email) {
      sessionContact.nickname = contact.nickname;
      sessionContact.status = contact.status;
      sessionContact.clientID = contact.clientID;
    }
  }
  contactsChanged();
}

MSNSession.prototype.getWindow = function() {
  return window;
}

MSNSession.prototype.getContactCount = function() {
  return this.contacts.length;
}

MSNSession.prototype.getFirstContactID = function() {
  return this.contacts[0].email;
}

MSNSession.prototype.getContact = function(email) {
  var contact = null;
  for (var i in this.contacts) {
    if (this.contacts[i].email == email) {
      contact = this.contacts[i];
    }
  }
  return contact;
}

MSNSession.prototype.getContactIndex = function(email) {
  var index;
  for (var i in this.contacts) {
    if (this.contacts[i].email == email) {
      index = i;
    }
  }
  return index;
}